import abc

from langchain_core.language_models import BaseChatModel


class Prompt:
    def __init__(self, prompt: str, ext_context: str = ''):
        self.seperator = f'\n{'-'*5}You are my assistant{'-'*5}\n'
        self.ext_context = ext_context
        self.prompt = f'{self.ext_context}{self.seperator}{prompt}'

    @abc.abstractmethod
    def invoke(self, model: BaseChatModel):
        pass



import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class IntrospectionPrompt(Prompt):
    def __init__(self, query: str, prev_results: list, ext_context: str = ''):
        prompt = ('Precondition: Below are actions you(the bot/assistant) have performed to achieve the user query. '
                  'You are talking to the user, use "you" instead of the "user", '
                  'and you are the assistant.\n'
                  f'User query: "{query}"\n'
                  "Task: tell user's intention first, "
                  "then think seriously whether you have achieve user's query based on actions you have performed. "
                  "Finally, provide the results wanted by user according to [user query]. "
                  "(If you did not achieve user's query, explain why?)\n"
                  'Output format: text.\n'
                  'Example output: Your intention is to calculate math, and I was trying to open calculator. '
                  'But i failed because i did not have that ability to open calculator. '
                  'I have not achieve your intention. '
                  '(Your output should be concise and comprehensive)\n'
                  f'Actions You Have Performed (from left to right): {prev_results}\n')
        super().__init__(prompt, ext_context)
        log.debug(f'IntrospectionPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        return model.invoke(self.prompt).content



from .prompt import Prompt
from .analyser_prompt import AnalyserPrompt
from .scheduler_prompt import SchedulerPrompt
from .executor_prompt import ExecutorPrompt
from .introspection_prompt import IntrospectionPrompt
from .query_resolver_prompt import QueryResolverPrompt



import logging

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class SchedulerPrompt(Prompt):
    def __init__(self, query: str, actions: list[Action], ext_context: str = ''):
        self.actions = actions
        prompt = dict()
        for action in self.actions:
            prompt[action.name] = str(action)
        prompt = ('Precondition: Below are the tools you can use (you can only use the following tools). '
                  f'Now there is a user query: "{query}"\n'
                  'Task: What you need to do is to plan your workflow based on the tools and user query.\n'
                  '(User query might contains many steps, '
                  'think carefully about every step and plan your workflow based on your tools)\n'
                  "(User's query might need to use one tool more than once, "
                  'but think carefully about how many times a tool needs to be used based on practical query, '
                  'do not abuse or overuse!)\n'
                  "(Sometimes some of the tools are irrelevant to user's query. Make sure to use tools properly.)\n"
                  'Output format: [{tool1.name}, {tool2.name}, ...] sequential and well-organized with no additional redundant information\n'
                  'Example output: [do_step_one, do_step_two, do_step_three]\n'
                  f'Tools: {prompt}\n')
        super().__init__(prompt, ext_context)
        log.debug(f'SchedulerPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> list[Action]:
        results = model.invoke(self.prompt)
        results = results.content[1:-1].split(',')
        _fin_results = list()
        for _a_result in results:
            for action in self.actions:
                if action.name == _a_result.strip():
                    _fin_results.append(action)
        return _fin_results



import langchain_openai
from langchain_core.language_models import BaseChatModel

DEFAULT_TMP = 0.5
DEFAULT_TOP_P = 0.5
DEFAULT_GPT = 'gpt-4o-mini'


def get_lm(key: str = None, name: str = DEFAULT_GPT, temp: float = DEFAULT_TMP, top_p: float = DEFAULT_TOP_P,
           stream: bool = False) -> BaseChatModel:
    return langchain_openai.ChatOpenAI(
        api_key=key,
        model=name,
        top_p=top_p,
        temperature=temp,
        streaming=stream
    )



import json
import logging

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class AnalyserPrompt(Prompt):
    def __init__(self, query: str, prev_results: list, action: Action, ext_context: str = ''):
        self.action = action
        prev_results_str = str()
        prompt = str()
        if len(prev_results) > 0:
            prev_results_str += '\n'
        for result in prev_results:
            prev_results_str += f'{result};\n'
        parameters_str = '{'
        for key, clazz in action.parameters.items():
            parameters_str += f'"{key}"(paramName):"{clazz}"(paramDataType),'
        parameters_str += '}'
        prompt += ('Precondition: Below is the tool you can use (you can only use this tool). '
                   f'Now there is a user query: "{query}"\n'
                   'Task: What you need to do is to generate values of parameters of the function\n'
                   'Output format: {param1_name:param1_value, param2_name:param2_value, ...} '
                   '(Parameters are given below at [!!!Parameters provided!!!], values are gonna be generated by you, '
                   'strictly abide by the parameter name and type)\n'
                   "(You must not generate any nonexistent parameter, "
                   "look carefully at parameters which are provided below at [!!!Parameters provided!!!]!)\n"
                   'Example output: {"param1": "value1", "param2": "value2"} '
                   '(With not any redundant information, only a json)\n'
                   f'Tool: {action}\n'
                   f'!!!Parameters provided!!!: {parameters_str}\n')
        if len(prev_results_str) != 0:
            prompt += (f'Generate response according to previous actions performed by you.\n'
                       f'Previous actions: {prev_results_str}\n'
                       f'What [previous actions] has shown is what you have done until now. '
                       f'What you need to do now is to think seriously what you have not done yet '
                       f'according to the [user query].\n')
        else:
            prompt += 'This is your first step, no previous actions have been performed.\n'
        super().__init__(prompt, ext_context)
        log.debug(f'AnalyserPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> tuple[Action, dict]:
        result = model.invoke(self.prompt).content
        param = json.loads(result)
        return self.action, param



import logging
from collections.abc import Iterator

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class ExecutorPrompt(Prompt):
    def __init__(self, params: dict, action: Action, ext_context: str = ''):
        self.action = action
        self.params = params
        prompt = ('Precondition: Below is a tool and your choice(params) for the tool.\n'
                  'Task: Explain what you are going to do.\n'
                  'Output format: text.\n'
                  'Example output: I am trying to open calculator.\n'
                  f'Tool: {action}\n'
                  f'Params(choice): {params}\n')
        super().__init__(prompt, ext_context)
        log.debug(f'Executor prompt(before): {self.prompt}')

    def explain(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        if stream:
            return model.stream(self.prompt)
        return model.invoke(self.prompt).content

    def invoke(self, model: BaseChatModel, stream: bool = False) -> str | Iterator:
        result = self.action.function(**self.params)
        prompt = ('Precondition: Below is a tool, your choice(params) for the tool, '
                  'and the result of your using the tool.\n'
                  'Task: Explain what you have done, and write down the result detailed. '
                  '(The result is shown below at [Result])\n'
                  'Output format: text.\n'
                  'Output contains: {the_choice_you_made}, {what_you_have_done}\n'
                  'Output example: I wrote a wechat message which says "Bonjour". The result is "success"\n'
                  f'Tool: {self.action}\n'
                  f'Params(choice): {self.params}\n'
                  f'Result: {result}\n')
        prompt = f'{self.ext_context}{self.seperator}{prompt}'
        log.debug(f'Executor prompt(after): {prompt}')
        if stream:
            return model.stream(prompt)
        return model.invoke(prompt).content



import json
from typing import Callable


class Action:
    def __init__(self, function: Callable):
        self.name: str = function.__name__
        self.description: str = function.__doc__
        self.function: Callable = function
        self.parameters: dict = dict()
        for key, clazz in function.__annotations__.items():
            if key != 'return':
                self.parameters[key] = clazz.__name__

    def __repr__(self):
        return json.dumps(obj={
            'name': self.name,
            'description': self.description,
            'parameters': self.parameters
        }, ensure_ascii=False)

    def __str__(self):
        return self.__repr__()



import logging
from typing import Callable

from langchain_core.language_models import BaseChatModel

from ceo.action.action import Action
from ceo.prompt import (
    SchedulerPrompt,
    AnalyserPrompt,
    ExecutorPrompt,
    IntrospectionPrompt,
    QueryResolverPrompt
)

log = logging.getLogger('ceo')


class Agent:
    def __init__(self, functions: list[Callable], model: BaseChatModel, query: str = '', ext_context: str = ''):
        self.actions = list()
        self.prev_results = list()
        self.schedule = list()
        self.act_count = 0
        self.model = model
        self.ext_context = ext_context
        self.query_high_level, self.query_by_step = (
            QueryResolverPrompt(query=query, ext_context=ext_context).invoke(self.model)
        )
        for function in functions:
            self.actions.append(Action(function))

    def plan(self) -> list:
        scheduling = SchedulerPrompt(query=self.query_by_step, actions=self.actions, ext_context=self.ext_context)
        self.schedule = scheduling.invoke(self.model)
        log.debug(f'Schedule: {[_.name for _ in self.schedule]}. Query: "{self.query_high_level}".')
        return self.schedule

    def reposition(self):
        self.prev_results = list()
        self.schedule = list()
        self.act_count = 0
        return self

    def assign(self, query: str):
        self.query_high_level, self.query_by_step = (
            QueryResolverPrompt(query=query, ext_context=self.ext_context).invoke(self.model))
        return self.reposition()

    def reassign(self, query: str):
        return self.assign(query=query)

    def step_quiet(self) -> str:
        if self.act_count < len(self.schedule):
            analysing = AnalyserPrompt(
                query=self.query_by_step,
                prev_results=self.prev_results,
                action=self.schedule[self.act_count],
                ext_context=self.ext_context
            )
            action, params = analysing.invoke(self.model)
            executing = ExecutorPrompt(params=params, action=action, ext_context=self.ext_context)
            action_str = f'Action {self.act_count + 1}/{len(self.schedule)}: {executing.invoke(model=self.model)}'
            self.prev_results.append(action_str)
            self.act_count += 1
            log.debug(action_str)
            return action_str
        self.reposition()
        return ''

    def just_do_it(self) -> str | None:
        if not self.plan():
            return None
        for act_count in range(len(self.schedule)):
            self.step_quiet()
        response = (IntrospectionPrompt(
            query=self.query_high_level,
            prev_results=self.prev_results,
            ext_context=self.ext_context).invoke(self.model))
        log.debug(f'Conclusion: {response}')
        self.reposition()
        return response



import logging

from langchain_core.language_models import BaseChatModel

from ceo.prompt.prompt import Prompt

log = logging.getLogger('ceo.prompt')


class QueryResolverPrompt(Prompt):
    def __init__(self, query: str, ext_context: str = ''):
        prompt = (f'Precondition: There is a user query: "{query}"\n'
                  "Task: What you need to do is to tell user's intention based on [user query]. \n"
                  "Task Redeclare: To tell user's intention based on [user query]. "
                  "Not your (you are the assistant) intention.\n"
                  "(Firstly, deduce the user's query step by step; "
                  "Secondly break user's intention down into several minimum steps;)\n"
                  'Output format: Step[n]:[Action of the step]\n'
                  'Example output: Step1:Open the door;Step2:Go into the room;Step3:Find the toys in the room;\n')
        self.__query = query
        super().__init__(prompt, ext_context)
        log.debug(f'QueryResolverPrompt: {self.prompt}')

    def invoke(self, model: BaseChatModel) -> tuple[str, str]:
        if self.__query == '':
            return f"User's intention: Don't do anything.", f"User's query(Step by step): Don't do anything."
        user_query_by_step = model.invoke(self.prompt).content
        summary_prompt = ("Task: Summarize user's query into a short sentence "
                          "which includes all the key information from user's query"
                          "(User's query is provided below at [User's query])\n"
                          f"User's query: \"{user_query_by_step}\"\n"
                          "Output format: string(summarization of [User's query])\n"
                          "Output example: To find toys for you in the room.\n")
        summary = model.invoke(summary_prompt).content
        return f"User's intention: {summary}", f"User's query(Step by step): {user_query_by_step}"



import setuptools

with open("README.md", "r", encoding="utf-8") as fh:
    long_description = fh.read()

with open("LICENSE", "r", encoding="utf-8") as fh:
    license_content = fh.read()

setuptools.setup(
    name="ceo-py",
    version='0.5.5-beta',
    author="vortezwohl",
    author_email="vortez.wohl@gmail.com",
    description="CEO is an easy-to-use AI agent framework. Now, you have taken on the role of the chairman of the board, instruct your CEO to carry out the task for you.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    license=license_content,
    url="https://github.com/vortezwohl/CEO",
    project_urls={
        "Bug Tracker": "https://github.com/vortezwohl/CEO/issues",
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "Operating System :: OS Independent",
        "Natural Language :: English",
    ],
    package_dir={"": "."},
    packages=setuptools.find_packages(where="."),
    python_requires=">=3.10",
    install_requires=[
        'langchain-core>=0.3.13',
        'langchain-openai>=0.2.3'
    ],
    entry_points={},
    include_package_data=False
)


import logging

logger = logging.getLogger('ceo')
logger.setLevel(logging.INFO)
console_handler = logging.StreamHandler()
console_handler.setLevel(logging.DEBUG)
formatter = logging.Formatter('[%(levelname)s] %(asctime)s %(name)s : %(message)s')
console_handler.setFormatter(formatter)
logger.addHandler(console_handler)

logger = logging.getLogger('ceo.prompt')
logger.setLevel(logging.INFO)
